from annet.annlib.rulebook import common
from annet.annlib.types import Op


def default_diff(old, new, diff_pre, _pops=(Op.AFFECTED,)):
    diff = common.base_diff(old, new, diff_pre, _pops, moved_to_affected=True)
    diff[:] = _skip_non_ap_env_affected(diff)
    return diff


def _skip_non_ap_env_affected(diff):
    for x in diff:
        if x.op == Op.AFFECTED and not x.children:
            if x.diff_pre["attrs"]["context"].get("block") != "ap-env":
                continue
        yield x



from dataclasses import dataclass
from datetime import datetime
from typing import List, Optional, Any, Dict

from annet.adapters.netbox.common.models import Entity, DeviceType


@dataclass
class Label:
    value: int
    label: str


@dataclass
class DeviceIp:
    id: int
    address: str
    family: int


@dataclass
class Device(Entity):
    url: str
    display_name: str
    device_type: DeviceType
    device_role: Entity
    tenant: Optional[Entity]
    platform: Optional[Entity]
    serial: str
    asset_tag: Optional[str]
    site: Entity
    rack: Optional[Entity]
    position: Optional[float]
    face: Optional[Label]
    status: Label
    primary_ip: Optional[DeviceIp]
    primary_ip4: Optional[DeviceIp]
    primary_ip6: Optional[DeviceIp]
    tags: List[str]
    custom_fields: Dict[str, Any]
    created: datetime
    last_updated: datetime


@dataclass
class Interface(Entity):
    device: Entity
    enabled: bool


@dataclass
class Vrf(Entity):
    rd: str


@dataclass
class IpAddress:
    id: int
    family: int
    address: str
    vrf: Optional[Vrf]
    tenant: Any  # ???
    status: Label
    description: Optional[str]
    custom_fields: Dict[str, Any]
    tags: List[str]
    created: datetime
    last_updated: datetime

    interface: Entity

    nat_inside: Any  # ???
    nat_outside: Any  # ???



from annet.generators import PartialGenerator


class InitialConfig(PartialGenerator):

    """
    Конфиги у свежих (еще ни разу не настраиваемых устройств)
    на самом деле НЕ пустые. В данном генераторе отображен
    такой набор команд, по крайней мере тех, которые могут
    изменяться в ходе первичной конфигурации.

    Acl для данного генератора не нужен, он будет генерировать
    конфиг целиком.
    """
    def __init__(self, storage=None):
        self._do_run = not storage
        super().__init__(storage=storage)

    def run_huawei(self, device):
        if not self._do_run:
            return
        if device.hw.CE:
            yield """
            telnet server disable
            telnet ipv6 server disable
            diffserv domain default
            aaa
                authentication-scheme default
                authorization-scheme default
                accounting-scheme default
                domain default
                domain default_admin
            """



from collections import defaultdict

from annet.annlib.types import Op

from annet.rulebook.common import default, default_diff


def user(key, diff, **_):
    check_for_remove = True
    added = []
    for add in diff[Op.ADDED]:
        added.append((True, add["row"], None))
        if add["row"].startswith("local-user %s password" % key[0]):
            check_for_remove = False
    if check_for_remove:
        for rem in diff[Op.REMOVED]:
            # Обрабатывать удаление только пароля, если меняется что-то другое, можно просто накатить без удаления
            if rem["row"].startswith("local-user %s password" % key[0]):
                yield (False, "undo local-user %s" % key[0], None)
                return
    yield from added


def domain(rule, key, diff, **_):
    """
    При удалении метода для accounting|authorization|authentication
    не нужно указывать сам метод, поэтому откидываем последний ключ.
    """
    if diff[Op.REMOVED]:
        yield (False, rule["reverse"].format(key[0], ""), None)
    else:
        yield from default(rule, key, diff)


def local_user_diff(old, new, diff_pre, **kwargs):
    diff = default_diff(old, new, diff_pre, **kwargs)
    filtered_diff = []
    # Группируем команды local-user по пользователю
    # и назначению (mode будет "password", "service-type", etc.)
    # {("username", "mode"): {op: diff_item}}}
    grouped = defaultdict(dict)
    for diff_item in diff:
        username, mode = _local_user_row_key(diff_item.row)
        if username and mode:
            grouped[(username, mode)][diff_item.op] = diff_item

    filtered_diff = []
    for diff_item in diff:
        username, mode = _local_user_row_key(diff_item.row)
        if username and mode:
            ops = set(grouped[(username, mode)])
            # NOCDEVDUTY-1786 делаем так чтобы в генераторе не требовалось точно попасть в порядок service-type
            # у хуавей порядок аргументов в данном месте меняется в зависимости от версии софта
            # при этом команда принимается в любом виде, меняется отображение в конфиге, вводить ее можно как угодно
            # поэтому если команды local-user * service-type ... совпадают с точностью до перестановки то ничего не правим
            if mode == "service-type" and ops == {Op.ADDED, Op.REMOVED}:
                added = set(grouped[(username, mode)][Op.ADDED].row.split())
                removed = set(grouped[(username, mode)][Op.REMOVED].row.split())
                if added == removed:
                    continue

        filtered_diff.append(diff_item)

    return filtered_diff


def _local_user_row_key(row):
    username, mode = None, None
    splitted_row = row.split()
    # Ожидаемый формат команды 'local-user <username> <mode> ...'
    if splitted_row and splitted_row[0] == "local-user":
        if len(splitted_row) >= 3:
            username = splitted_row[1]
            mode = splitted_row[2]
    return username, mode



"""Support JSON patch (RFC 6902) and JSON Pointer (RFC 6901)."""

import copy
import json
from typing import Any, Dict, List, Optional

import jsonpatch
import jsonpointer


def format_json(data: Any, stable: bool = False) -> str:
    """Serialize to json."""
    return json.dumps(data, indent=4, ensure_ascii=False, sort_keys=not stable) + "\n"


def apply_json_fragment(
        old: Dict[str, Any],
        new_fragment: Dict[str, Any],
        acl: List[str],
) -> Dict[str, Any]:
    """
    Replace parts of the old document with 'new_fragment' using ACL restrictions.
    """
    full_new_config = copy.deepcopy(old)
    for acl_item in acl:
        pointer = jsonpointer.JsonPointer(acl_item)

        try:
            new_value = pointer.get(new_fragment)
        except jsonpointer.JsonPointerException:
            # no value found in new_fragment by the pointer, skip the ACL item
            continue

        _ensure_pointer_exists(full_new_config, pointer)
        pointer.set(full_new_config, new_value)

    return full_new_config


def _ensure_pointer_exists(doc: Dict[str, Any], pointer: jsonpointer.JsonPointer) -> None:
    """
    Ensure that document has all pointer parts (if possible).

    This is workaround for errors of type:

    ```
    jsonpointer.JsonPointerException: member 'MY_PART' not found in {}
    ```

    See for details: https://github.com/stefankoegl/python-json-pointer/issues/41
    """
    parts_except_the_last = pointer.get_parts()[:-1]
    doc_pointer: Dict[str, Any] = doc
    for part in parts_except_the_last:
        if part not in doc_pointer:
            # create an empty object by the pointer part
            doc_pointer[part] = {}

        if isinstance(doc_pointer, dict):
            # follow the pointer to delve deeper
            doc_pointer = doc_pointer[part]
        else:
            # not a dict - cannot delve deeper
            break


def make_patch(old: Dict[str, Any], new: Dict[str, Any]) -> List[Dict[str, Any]]:
    """Generate a JSON patch by comparing the old document with the new one."""
    return jsonpatch.make_patch(old, new).patch


def apply_patch(content: Optional[bytes], patch_bytes: bytes) -> bytes:
    """
    Apply JSON patch to file contents.

    If content is None it is considered that the file does not exist.
    """
    old_doc: Any
    if content is not None:
        old_doc = json.loads(content)
    else:
        old_doc = None

    patch_data = json.loads(patch_bytes)
    patch = jsonpatch.JsonPatch(patch_data)
    new_doc = patch.apply(old_doc)

    new_contents = format_json(new_doc, stable=True).encode()
    return new_contents


def apply_acl_filters(content: Dict[str, Any], filters: List[str]) -> Dict[str, Any]:
    result = {}
    for f in filters:
        filter_text = f.strip()
        if not filter_text:
            continue

        pointer = jsonpointer.JsonPointer(filter_text)

        try:
            part = pointer.get(copy.deepcopy(content))

            sub_tree = result
            for i in pointer.get_parts():
                if i not in sub_tree:
                    sub_tree[i] = {}
                sub_tree = sub_tree[i]

            patch = jsonpatch.JsonPatch([{"op": "add", "path": filter_text, "value": part}])
            result = patch.apply(result)
        except jsonpointer.JsonPointerException:
            # no value found in content by the pointer, skip the ACL item
            continue

    return result



import itertools
import re
from collections import OrderedDict as odict

from .types import Op


# =====
class ParserError(Exception):
    pass


# =====
class _CommentOrEmpty:
    pass


class BlockBegin:
    pass


class BlockEnd:
    pass


# =====
class CommonFormatter:
    def __init__(self, indent="  "):
        self._indent = indent
        self._block_begin = ""
        self._block_end = ""
        self._statement_end = ""

    def split(self, text):
        return list(filter(None, text.split("\n")))

    def join(self, config):
        return "\n".join(
            _filtered_block_marks(
                self._indent_blocks(self._blocks(config, is_patch=False))
            )
        )

    def diff(self, diff):
        return list(self._diff_lines(diff))

    def diff_generator(self, diff):
        yield from self._diff_lines(diff)

    def patch(self, patch):
        return "\n".join(
            _filtered_block_marks(
                self._indent_blocks(self._blocks(patch, is_patch=True))
            )
        )

    def cmd_paths(self, patch):
        ret = odict()
        path = []
        for row, context in self.blocks_and_context(patch, is_patch=True):
            if row is BlockBegin:
                path.append(path[-1])
            elif row is BlockEnd:
                path.pop()
            else:
                if path:
                    path.pop()
                path.append(row)
                ret[tuple(path)] = context
        return ret

    def patch_plain(self, patch):
        return list(self.cmd_paths(patch).keys())

    def _diff_lines(self, diff, _level=0, _block_sign=None):
        sign_map = {
            Op.REMOVED: "-",
            Op.ADDED: "+",
            Op.MOVED: ">",
            Op.AFFECTED: " ",
        }
        for (flag, row, children, _) in diff:
            sign = sign_map[flag]
            if not children:
                yield "%s %s%s" % (sign, self._indent * _level, row + self._statement_end)
            else:
                yield "%s %s%s" % (sign, self._indent * _level, row + self._block_begin)
                yield from self._diff_lines(children, _level + 1, sign)
        if _level > 0 and self._block_end and _block_sign is not None:
            yield "%s %s%s" % (_block_sign, self._indent * (_level - 1), self._block_end)

    def _indented_blocks(self, tree):
        return self._indent_blocks(self._blocks(tree, False))

    def _indent_blocks(self, blocks):
        _level = 0
        for row in blocks:
            if row is BlockBegin:
                _level += 1
            elif row is BlockEnd:
                _level -= 1
            else:
                row = self._indent * _level + row
            yield row

    def blocks_and_context(self, tree, is_patch, _level=0, _parent=None):
        empty_context = {}
        if is_patch:
            items = ((item.row, item.child, item.context) for item in tree.itms)
        else:
            items = ((row, child, empty_context) for row, child in tree.items())
        for row, sub_config, context in items:
            yield row, context
            if sub_config or (is_patch and sub_config is not None):
                yield BlockBegin, None
                yield from self.blocks_and_context(sub_config, is_patch, _level + 1, row)
                yield BlockEnd, None

    def _blocks(self, tree, is_patch):
        for row, _context in self.blocks_and_context(tree, is_patch):
            yield row


class BlockExitFormatter(CommonFormatter):
    def __init__(self, block_exit, no_block_exit=(), indent="  "):
        super().__init__(indent)
        self._block_exit = block_exit
        self._no_block_exit = tuple(no_block_exit)

    def split_remove_spaces(self, text):
        # эта регулярка заменяет 2 и более пробела на один, но оставляет пробелы в начале линии
        text = re.sub(r"(?<=\S)\ {2,}(?=\S)", " ", text)
        res = super().split(text)
        return res

    def block_exit(self, parent):
        if not (parent and parent.startswith(self._no_block_exit)):
            return self._block_exit

    def blocks_and_context(self, tree, is_patch, _level=0, _parent=None):
        last_context = {}
        for row, context in super().blocks_and_context(tree, is_patch, _level, _parent):
            yield row, context
            if context is not None:
                last_context = context
        if is_patch and _level > 0:
            exit_statement = self.block_exit(_parent)
            if exit_statement:
                yield exit_statement, last_context


class HuaweiFormatter(BlockExitFormatter):
    def __init__(self, indent="  "):
        super().__init__(
            block_exit="quit",
            no_block_exit=[
                "rsa peer-public-key",
                "dsa peer-public-key",
                "public-key-code begin",
            ],
            indent=indent,
        )

    def split(self, text):
        # на старых прошивка наблюдается баг с двумя пробелами в этом месте в конфиге
        # например на VRP V100R006C00SPC500 + V100R006SPH003
        policy_end_blocks = ("end-list", "endif", "end-filter")
        tree = self.split_remove_spaces(text)
        tree[:] = filter(lambda x: not x.endswith(policy_end_blocks), tree)
        return tree

    def block_exit(self, parent):
        if parent:
            if parent.startswith("xpl route-filter"):
                return "end-filter"
            elif parent.startswith(("xpl")):
                return "end-list"
            elif parent.startswith("if") and parent.endswith("then"):
                return "endif"
        return super().block_exit(parent)


class CiscoFormatter(BlockExitFormatter):
    def __init__(self, indent="  "):
        super().__init__("exit", indent)

    def split(self, text):
        return self.split_remove_spaces(text)


class AsrFormatter(BlockExitFormatter):
    def __init__(self, indent="  "):
        super().__init__("exit", indent)

    def split(self, text):
        policy_end_blocks = ("end-set", "endif", "end-policy")
        tree = self.split_remove_spaces(text)
        tree[:] = filter(lambda x: not x.endswith(policy_end_blocks), tree)
        return tree

    def block_exit(self, parent):
        if parent:
            if parent.startswith(("prefix-set", "as-path-set", "community-set")):
                return "end-set"
            elif parent.startswith("if") and parent.endswith("then"):
                return "endif"
            elif parent.startswith("route-policy"):
                return "end-policy"
        return super().block_exit(parent)


class JuniperFormatter(CommonFormatter):
    patch_set_prefix = "set "

    def __init__(self, indent="    "):
        super().__init__(indent)
        self._block_begin = " {"
        self._block_end = "}"
        self._statement_end = ";"
        self._endofline_comment = "; ##"

    def split(self, text):
        sub_regexs = (
            (re.compile(self._block_begin + r"\s*" + self._block_end + r"$"), ""),  # collapse empty blocks
            (re.compile(self._block_begin + "(\t# .+)?$"), ""),
            (re.compile(self._statement_end + r"$"), ""),
            (re.compile(r"\s*" + self._block_end + "(\t# .+)?$"), ""),
            (re.compile(self._endofline_comment + r".*$"), ""),
        )
        split = []
        for line in text.split("\n"):
            for (regex, repl_line) in sub_regexs:
                line = regex.sub(repl_line, line)
            split.append(line)
        return list(filter(None, split))

    def join(self, config):
        return "\n".join(_filtered_block_marks(self._formatted_blocks(self._indented_blocks(config))))

    def patch(self, patch):
        return "\n".join(" ".join(x) for x in self.cmd_paths(patch))

    def patch_plain(self, patch):
        return list(self.cmd_paths(patch).keys())

    def _formatted_blocks(self, blocks):
        level = 0
        line = None
        for new_line in blocks:
            if new_line is BlockBegin:
                level += 1
                if isinstance(line, str):
                    yield line + self._block_begin
            elif new_line is BlockEnd:
                level -= 1
                if isinstance(line, str):
                    yield line + self._statement_end
                yield self._indent * level + self._block_end
            elif isinstance(line, str):
                yield line + self._statement_end
            line = new_line
        if isinstance(line, str):
            yield line + self._statement_end

    def cmd_paths(self, patch, _prev=""):
        commands = odict()
        for item in patch.itms:
            key, childs, context = item.row, item.child, item.context
            if childs:
                for k, v in self.cmd_paths(childs, _prev + " " + key).items():
                    commands[k] = v
            else:
                if key.startswith("delete"):
                    cmd = "delete" + _prev + " " + key.replace("delete", "", 1).strip()
                elif key.startswith("activate"):
                    cmd = "activate" + _prev + " " + key.replace("activate", "", 1).strip()
                elif key.startswith("deactivate"):
                    cmd = "deactivate" + _prev + " " + key.replace("deactivate", "", 1).strip()
                else:
                    cmd = (self.patch_set_prefix + _prev.strip()).strip() + " " + key
                # Expanding [ a b c ] junipers list of arguments
                matches = re.search(r"^(.*)\s+\[(.+)\]$", cmd)
                if matches:
                    for c in matches.group(2).split(" "):
                        if c.strip():
                            cmd = " ".join([matches.group(1), c])
                            commands[(cmd,)] = context
                else:
                    commands[(cmd,)] = context

        return commands


class RibbonFormatter(JuniperFormatter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._endofline_comment = "; # SECRET-DATA"


class JuniperList:
    """
    Форматирует inline-листы в конфиге juniper
    """

    def __init__(self, *args, spaces=True, **kwargs):
        self._items = list(*args, **kwargs)
        self.spaces = spaces

    def __str__(self):
        if self.spaces:
            return "[ %s ]" % " ".join(str(_) for _ in self._items)
        else:
            return "[%s]" % " ".join(str(_) for _ in self._items)


class NokiaFormatter(JuniperFormatter):
    patch_set_prefix = "/configure "

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._statement_end = ""
        self._endofline_comment = " ##"

    def split(self, text):
        ret = super().split(text)
        # NOCDEVDUTY-248 сдергиваем верхний configure-блок
        # NOCDEVDUTY-282 после configure {} блока могут идти еще блоки которые нам не нужны
        start, finish = None, None
        for i, line in enumerate(ret):
            if line.startswith("#"):
                continue
            # начало configure-блока
            if line == "configure":
                start = i + 1
            # любой после configure последующий блок на глобальном уровне
            elif len(line) == len(line.lstrip()):
                if start is not None and finish is None:
                    finish = i
        # Если configure-блока не было - то весь конфиг считаем configre'ом
        start = start if start is not None else 0
        finish = finish if finish is not None else len(ret)
        return ret[start:finish]

    def cmd_paths(self, patch, _prev=""):
        commands = odict()
        for item in patch.itms:
            key, childs, context = item.row, item.child, item.context
            if childs:
                for k, v in self.cmd_paths(childs, _prev + " " + key).items():
                    commands[k] = v
            else:
                if key.startswith("delete"):
                    cmd = "/configure delete" + _prev + " " + key.replace("delete", "", 1).strip()
                else:
                    cmd = self.patch_set_prefix + _prev.strip() + " " + key
                # Expanding [ a b c ] junipers list of arguments
                matches = re.search(r"^(.*)\s+\[(.+)\]$", cmd)
                if matches:
                    for c in matches.group(2).split(" "):
                        if c.strip():
                            cmd = " ".join([matches.group(1), c])
                            commands[(cmd,)] = context
                else:
                    commands[(cmd,)] = context
        return commands


class RosFormatter(CommonFormatter):
    patch_set_prefix = "set "

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._block_begin = "/"

    def join(self, config):
        return "\n".join(_filtered_block_marks(self._formatted_blocks(self._indented_blocks(config))))

    def patch(self, patch):
        return "\n".join(" ".join(x) for x in self.cmd_paths(patch))

    def patch_plain(self, patch):
        return list(self.cmd_paths(patch).keys())

    def blocks_and_context(self, tree, is_patch, _level=0, _parent=None):
        rows = []
        empty_context = {}
        if is_patch:
            items = ((item.row, item.child, item.context) for item in tree.itms)
        else:
            items = ((row, child, empty_context) for row, child in tree.items())
        for row, sub_config, context in items:
            if sub_config or (is_patch and sub_config is not None):
                rows.append((row, sub_config, context))
            else:
                rows.append((row, None, context))

        prev_prow = None
        for sub_config, row_group in itertools.groupby(rows, lambda x: x[1]):
            if sub_config is None:
                if prev_prow:
                    yield prev_prow
                    yield BlockBegin, None
                for (row, _, context) in row_group:
                    yield row, context
                if prev_prow:
                    yield BlockEnd, None
            else:
                for (row, _, context) in row_group:
                    if _parent:
                        prev_prow = _parent
                        prow = f"{_parent} {row}"
                    else:
                        prow = row
                    yield prow, context
                    yield BlockBegin, None
                    yield from self.blocks_and_context(sub_config, is_patch, _level + 1, prow)
                    yield BlockEnd, None

    def _formatted_blocks(self, blocks):
        line = None
        for new_line in blocks:
            if new_line is BlockBegin:
                if isinstance(line, str):
                    yield self._block_begin + line.strip()
            elif isinstance(line, str):
                yield line
            line = new_line

    def _splitter_file(self, lines):
        filedesrc_re = re.compile(r"^\s+(?P<num>\d+)\s+name=\"(?P<name>[^\"]+)\"\s+type=\"(?P<type>[^\"]+)\""
                                  r"\s+(size=(?P<size>.*))?creation-time=(?P<time>.*?)(contents=(?P<content>.*)?)?$")
        file_content_indent = re.compile(r"^\s{5}")
        out = []
        files = {}
        curfile = None
        for line in lines:
            match = filedesrc_re.search(line)
            if match:
                if match.group("type").strip() == ".txt file":
                    curfile = match.group("name")
                    files[curfile] = {"name": curfile, "contents": []}
                    if match.group("content"):
                        files[curfile]["contents"].append(match.group("content").strip())
            elif curfile and file_content_indent.match(line):
                files[curfile]["contents"].append(file_content_indent.sub("", line))
        for file in files.values():
            out.append(f"print file={file['name']}")
            if len(file["contents"]) > 0:
                text = "\n".join(file["contents"])
                out.append(f"set {file['name']} contents=\"{text}\"")
        return out

    def _splitter_user_ssh_keys(self, lines):
        keydescr_re = re.compile(r"user=(?P<user>\w+).*key-owner=(?P<owner>.*)$")
        out = []
        for line in lines:
            match = keydescr_re.search(line)
            if match:
                out.append(f"import public-key-file={match.group('owner')}.ssh_key.txt user={match.group('user')}")

        return out

    def split(self, text):
        split = []
        level = 0
        postj = {}
        curgroup = None
        for line in text.split("\n"):
            if line.startswith("/"):
                if curgroup:
                    for row in getattr(self, curgroup)(postj[curgroup]):
                        if level > 0:
                            row = row.strip()
                        split.append(self._indent * level + row)

                level = 0
                for group in line.split():
                    split.append(self._indent * level + group.replace("/", ""))
                    level += 1

                gpath = line.replace("/", "_splitter_").replace(" ", "_").replace("-", "_")
                if hasattr(self, gpath):
                    postj[gpath] = []
                    curgroup = gpath
                else:
                    curgroup = None
            else:
                row = line
                if curgroup:
                    postj[curgroup].append(row)
                else:
                    if level > 0:
                        row = line.strip()
                    split.append(self._indent * level + row)
        if curgroup:
            for row in getattr(self, curgroup)(postj[curgroup]):
                if level > 0:
                    row = row.strip()
                split.append(self._indent * level + row)
        return list(filter(None, split))

    def cmd_paths(self, patch, _prev=""):
        rm_regexs = (
            (re.compile(r"^add "), ""),
            (re.compile(r"^print file="), "name="),
        )
        patch_items = []
        for item in patch.itms:
            key, childs, context = item.row, item.child, item.context
            if childs:
                patch_items.append((key, childs, context))
            else:
                patch_items.append((key, None, context))

        commands = odict()
        prev_cmd = None
        prev_context = None
        for childs, items in itertools.groupby(patch_items, lambda x: x[1]):
            if childs is None:
                if prev_cmd:
                    commands[(prev_cmd,)] = prev_context
                for key, _, context in items:
                    if key.startswith("remove"):
                        find_cmd = key.replace("remove", "", 1).strip()
                        for (regex, repl_line) in rm_regexs:
                            find_cmd = regex.sub(repl_line, find_cmd)
                        cmd = "remove [ find " + find_cmd + " ]"
                    else:
                        cmd = key
                    commands[(cmd,)] = context
            else:
                for key, _, context in items:
                    if _prev:
                        prev_cmd = _prev
                        prev_context = context
                        block_cmd = f"{_prev} {key}"
                    else:
                        block_cmd = f"/{key}"
                    commands[(block_cmd,)] = context
                    for k, v in self.cmd_paths(childs, block_cmd).items():
                        commands[k] = v
        return commands


def make_formatter(vendor, **kwargs):
    formatters = {
        "juniper": JuniperFormatter,
        "cisco": CiscoFormatter,
        "nexus": CiscoFormatter,
        "huawei": HuaweiFormatter,
        "arista": CiscoFormatter,
        "nokia": NokiaFormatter,
        "routeros": RosFormatter,
        "aruba": CiscoFormatter,
        "pc": CommonFormatter,
        "ribbon": RibbonFormatter,
    }
    return formatters[vendor](**kwargs)


# ====
def parse_to_tree(text, splitter, comments=("!", "#")):
    tree = odict()
    for stack in _stacked(splitter(text), tuple(comments)):
        local_tree = tree
        for key in stack:
            if key not in local_tree:
                local_tree[key] = odict()
            local_tree = local_tree[key]
    return tree


# =====
def _stacked(lines, comments):
    stack = []
    for (level, line) in _stripped_indents(lines, comments):
        level += 1
        if level > len(stack):
            stack.append(line)
        elif level == len(stack):
            stack[-1] = line
        else:
            stack = stack[:level - 1] + [line]
        yield tuple(stack)


def _stripped_indents(lines, comments):
    indents = []
    curr_level = 0
    g_level = None

    for (number, (level, line)) in enumerate(_parsed_indents(lines, comments), start=1):
        if isinstance(line, str):
            if g_level is None:
                g_level = level
            level = level - (g_level or 0)
            if level < 0:
                raise ParserError("Invalid top indention: line %d: %s" % (number, line))

            if level > curr_level:
                indents.append(level - curr_level)
                curr_level += level - curr_level
            elif level < curr_level:
                while curr_level > level and len(indents):
                    curr_level -= indents.pop()
                if curr_level != level:
                    raise ParserError("Invalid top indention: line %d: %s" % (number, line))

            yield (len(indents), line)

        elif line is BlockEnd:
            indents = []
            curr_level = 0
            g_level = None


def _parsed_indents(lines, comments):
    for line in _filtered_lines(lines, comments):
        if isinstance(line, str):
            yield (_parse_indent(line), line.strip())
        else:
            yield (0, line)


def _filtered_lines(lines, comments):
    for line in lines:
        stripped = line.strip()
        # TODO Это для хуавей, так что хелпер нужно унести в Formatter
        if "#" in comments and line.startswith("#"):
            yield BlockEnd
        elif len(stripped) == 0 or stripped.startswith(comments):
            yield _CommentOrEmpty
        else:
            yield line


def _filtered_block_marks(blocks):
    return filter(lambda b: isinstance(b, str), blocks)


def _parse_indent(line):
    level = 0
    for ch in line:
        if ch in ("\t", " "):
            level += 1
        else:
            break
    return level



from __future__ import annotations

import abc
import contextlib
import textwrap
from typing import Union, List

from annet import tabparser, tracing
from annet.tracing import tracing_connector
from .exceptions import InvalidValueFromGenerator


class DefaultBlockIfCondition:
    pass


ParamsList = tabparser.JuniperList


class GenStringable(abc.ABC):
    @abc.abstractmethod
    def gen_str(self) -> str:
        pass


def _filter_str(value: Union[
    str, int, float, tabparser.JuniperList, ParamsList, GenStringable]):
    if isinstance(value, (
            str,
            int,
            float,
            tabparser.JuniperList,
            ParamsList,
    )):
        return str(value)

    if hasattr(value, "gen_str") and callable(value.gen_str):
        return value.gen_str()

    raise InvalidValueFromGenerator(
        "Invalid yield type: %s(%s)" % (type(value).__name__, value))


def _split_and_strip(text):
    if "\n" in text:
        rows = textwrap.dedent(text).strip().split("\n")
    else:
        rows = [text]
    return rows


# =====
class BaseGenerator:
    TYPE: str
    TAGS: List[str]

    def supports_device(self, device) -> bool:  # pylint: disable=unused-argument
        return True


class TreeGenerator(BaseGenerator):
    def __init__(self, indent="  "):
        self._indents = []
        self._rows = []
        self._block_path = []
        self._indent = indent

    @tracing.contextmanager(min_duration="0.1")
    @contextlib.contextmanager
    def block(self, *tokens, indent=None):
        span = tracing_connector.get().get_current_span()
        if span:
            span.set_attribute("tokens", " ".join(map(str, tokens)))

        indent = self._indent if indent is None else indent
        block = " ".join(map(_filter_str, tokens))
        self._block_path.append(block)
        self._append_text(block)
        self._indents.append(indent)
        yield
        self._indents.pop(-1)
        self._block_path.pop(-1)

    @contextlib.contextmanager
    def block_if(self, *tokens, condition=DefaultBlockIfCondition):
        if condition is DefaultBlockIfCondition:
            condition = (None not in tokens and "" not in tokens)
        if condition:
            with self.block(*tokens):
                yield
                return
        yield

    @contextlib.contextmanager
    def multiblock(self, *blocks):
        if blocks:
            blk = blocks[0]
            tokens = blk if isinstance(blk, (list, tuple)) else [blk]
            with self.block(*tokens):
                with self.multiblock(*blocks[1:]):
                    yield
                    return
        yield

    @contextlib.contextmanager
    def multiblock_if(self, *blocks, condition=DefaultBlockIfCondition):
        if condition is DefaultBlockIfCondition:
            condition = (None not in blocks)
            if condition:
                if blocks:
                    blk = blocks[0]
                    tokens = blk if isinstance(blk, (list, tuple)) else [blk]
                    with self.block(*tokens):
                        with self.multiblock(*blocks[1:]):
                            yield
                            return
        yield

    # ===
    def _append_text(self, text):
        self._append_text_cb(text)

    def _append_text_cb(self, text, row_cb=None):
        for row in _split_and_strip(text):
            if row_cb:
                row = row_cb(row)
            self._rows.append("".join(self._indents) + row)


class TextGenerator(TreeGenerator):
    def __add__(self, line):
        self._append_text(line)
        return self

    def __iter__(self):
        yield from self._rows



from .partial import PartialGenerator


class RefGenerator(PartialGenerator):
    def __init__(self, storage, groups=None):
        super().__init__(storage)
        self.groups = groups

    def ref(self, device):
        if hasattr(self, "ref_" + device.hw.vendor):
            return getattr(self, "ref_" + device.hw.vendor)(device)
        return ""

    def with_groups(self, groups):
        return type(self)(self.storage, groups)



from dataclasses import dataclass
from functools import wraps
from typing import Generic, Optional, List, TypeVar, Callable

from dataclass_rest.http.requests import RequestsClient
from requests import Session

Model = TypeVar("Model")


@dataclass
class PagingResponse(Generic[Model]):
    next: Optional[str]
    previous: Optional[str]
    count: int
    results: List[Model]


Func = TypeVar("Func", bound=Callable)


def _collect_by_pages(func: Func) -> Func:
    """Collect all results using only pagination."""
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        kwargs.setdefault("offset", 0)
        limit = kwargs.setdefault("limit", 100)
        results = []
        method = func.__get__(self, self.__class__)
        has_next = True
        while has_next:
            page = method(*args, **kwargs)
            kwargs["offset"] += limit
            results.extend(page.results)
            has_next = bool(page.next)
        return PagingResponse(
            previous=None,
            next=None,
            count=len(results),
            results=results,
        )

    return wrapper


# default batch size 100 is calculated to fit list of UUIDs in 4k URL length
def collect(func: Func, field: str = "", batch_size: int = 100) -> Func:
    """
    Collect data from method iterating over pages and filter batches.

    :param func: Method to call
    :param field: Field which defines a filter split into batches
    :param batch_size: Limit of values in `field` filter requested at a time
    """
    func = _collect_by_pages(func)
    if not field:
        return func

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        value = kwargs.get(field)
        if not value:
            return func(*args, **kwargs)

        method = func.__get__(self, self.__class__)
        results = []
        for offset in range(0, len(value), batch_size):
            kwargs[field] = value[offset:offset + batch_size]
            page = method(*args, **kwargs)
            results.extend(page.results)
        return PagingResponse(
            previous=None,
            next=None,
            count=len(results),
            results=results,
        )

    return wrapper


class BaseNetboxClient(RequestsClient):
    def __init__(self, url: str, token: str):
        url = url.rstrip("/") + "/api/"
        session = Session()
        if token:
            session.headers["Authorization"] = f"Token {token}"
        super().__init__(url, session)



from logging import getLogger
from typing import Optional, List, Union

from adaptix import P
from adaptix.conversion import impl_converter, link

from annet.adapters.netbox.common import models
from annet.adapters.netbox.common.manufacturer import (
    is_supported, get_hw, get_breed,
)
from annet.adapters.netbox.common.query import NetboxQuery
from annet.adapters.netbox.common.storage_opts import NetboxStorageOpts
from annet.annlib.netdev.views.hardware import HardwareView
from annet.storage import Storage
from . import api_models
from .client import NetboxV37

logger = getLogger(__name__)


@impl_converter(recipe=[
    link(P[api_models.Device].name, P[models.NetboxDevice].hostname),
    link(P[api_models.Device].name, P[models.NetboxDevice].fqdn),
])
def extend_device_base(
        device: api_models.Device,
        interfaces: List[models.Interface],
        hw: Optional[HardwareView],
        breed: str,
        storage: Storage,
        neighbours_ids: List[int],
) -> models.NetboxDevice:
    ...


def extend_device(
        device: api_models.Device, storage: Storage,
) -> models.NetboxDevice:
    return extend_device_base(
        device=device,
        interfaces=[],
        breed=get_breed(
            device.device_type.manufacturer.name,
            device.device_type.model,
        ),
        hw=get_hw(
            device.device_type.manufacturer.name,
            device.device_type.model,
        ),
        neighbours_ids=[],
        storage=storage,
    )


@impl_converter
def extend_interface(
        interface: api_models.Interface, ip_addresses: List[models.IpAddress],
) -> models.Interface:
    ...


class NetboxStorageV37(Storage):
    def __init__(self, opts: Optional[NetboxStorageOpts] = None):
        self.netbox = NetboxV37(
            url=opts.url,
            token=opts.token,
        )

    def __enter__(self):
        return self

    def __exit__(self, _, __, ___):
        pass

    def resolve_object_ids_by_query(self, query: NetboxQuery):
        return [
            d.id for d in self._load_devices(query)
        ]

    def resolve_fdnds_by_query(self, query: NetboxQuery):
        return [
            d.name for d in self._load_devices(query)
        ]

    def make_devices(
            self,
            query: Union[NetboxQuery, list],
            preload_neighbors=False,
            use_mesh=None,
            preload_extra_fields=False,
            **kwargs,
    ) -> List[models.NetboxDevice]:
        if isinstance(query, list):
            query = NetboxQuery.new(query)
        device_ids = {
            device.id: extend_device(device=device, storage=self)
            for device in self._load_devices(query)
        }
        if not device_ids:
            return []

        interfaces = self._load_interfaces(list(device_ids))
        for interface in interfaces:
            device_ids[interface.device.id].interfaces.append(interface)
        return list(device_ids.values())

    def _load_devices(self, query: NetboxQuery) -> List[api_models.Device]:
        if not query.globs:
            return []
        return [
            device
            for device in self.netbox.all_devices(
                name__ic=query.globs,
            ).results
            if _match_query(query, device)
            if is_supported(device.device_type.manufacturer.name)
        ]

    def _load_interfaces(self, device_ids: List[int]) -> List[
        models.Interface]:
        interfaces = self.netbox.all_interfaces(device_id=device_ids)
        extended_ifaces = {
            interface.id: extend_interface(interface, [])
            for interface in interfaces.results
        }

        ips = self.netbox.all_ip_addresses(interface_id=list(extended_ifaces))
        for ip in ips.results:
            extended_ifaces[ip.assigned_object_id].ip_addresses.append(ip)
        return list(extended_ifaces.values())

    def get_device(
            self, obj_id, preload_neighbors=False, use_mesh=None,
            **kwargs,
    ) -> models.NetboxDevice:
        device = self.netbox.get_device(obj_id)
        res = extend_device(device=device, storage=self)
        res.interfaces = self._load_interfaces([device.id])
        return res

    def flush_perf(self):
        pass


def _match_query(query: NetboxQuery, device_data: api_models.Device) -> bool:
    for subquery in query.globs:
        if subquery.strip() in device_data.name:
            return True
    return False



import functools
import ipaddress
from typing import Dict, Generator

import colorama

from .types import Diff, Op

# NOCDEV-1720


diff_ops = {
    "+": Op.ADDED,
    "-": Op.REMOVED,
    " ": Op.AFFECTED,
    ">": Op.MOVED,
}

ops_sign = {
    v: k
    for k, v in diff_ops.items()
}

ops_order = {
    Op.AFFECTED: 0,
    Op.MOVED: 1,
    Op.REMOVED: 2,
    Op.ADDED: 3,
}

ops_color = {
    Op.REMOVED: colorama.Fore.RED,
    Op.ADDED: colorama.Fore.GREEN,
    Op.AFFECTED: colorama.Fore.CYAN,
    Op.MOVED: colorama.Fore.YELLOW,
}


def is_int(ts):
    try:
        int(ts)
        return True
    except ValueError:
        return False


def is_ip(ts):
    try:
        ipaddress.ip_interface(ts)
        return True
    except ValueError:
        return False


def diff_cmp(diff_l, diff_r):
    """
Сборник костылей для сравнения двух строк диффа
    """
    (op_l, line_l, _, _) = diff_l
    (op_r, line_r, _, _) = diff_r

    cmp_line = (line_l > line_r) - (line_l < line_r)
    cmp_op = ops_order[op_l] - ops_order[op_r]
    if cmp_line == 0:
        # При равенстве строк порядок определяется операцией
        return cmp_op

    if cmp_op == 0:
        # Если для строк операции одинаковы, то считаем их равными
        # По идее TimSort стабилен и порядок просто не должен измениться
        return 0

    # Для частично совпадающих строк
    lws = line_l.split(" ")
    rws = line_r.split(" ")
    res = 0
    for i, lw in enumerate(lws):
        if len(rws) > i:
            rw = rws[i]
            if is_int(lw) and is_int(rw):
                # В одинаковом положении в строках есть инты, сортируем по ним
                res = int(lw) - int(rw)
                if res == 0:
                    # При равных интах - сортируем по операции
                    res = cmp_op
            elif is_ip(lw) and is_ip(rw):
                # Аналогично интам обрабатываем ip-адреса
                ip_l = ipaddress.ip_interface(lw)
                ip_r = ipaddress.ip_interface(rw)
                try:
                    res = (ip_l > ip_r) - (ip_l < ip_r)
                except TypeError:
                    res = 1
                    if ip_l.version == 4:
                        res = -1
                if res == 0:
                    res = cmp_op
            elif i > 0:
                if lw == rw:
                    res = cmp_op
            else:
                continue
            break
    if res != 0:
        return res
    return cmp_line


def resort_diff(diff: Diff) -> Diff:
    res = []
    df = sorted(diff, key=functools.cmp_to_key(diff_cmp))
    for line in df:
        ln = line
        if len(line[2]) > 0:
            ln = (line[0], line[1], resort_diff(line[2]), line[3])
        res.append(ln)
    return res


def colorize_line_with_color(line: str, color: int, no_color: bool):
    stripped = line.rstrip("\n")
    add_newlines = len(line) - len(stripped)
    line = stripped

    if not no_color:
        line = "%s%s%s%s" % (colorama.Style.BRIGHT, color, line, colorama.Style.RESET_ALL)

    line += "\n" * add_newlines
    return line


def colorize_line(line, no_color=False):
    op = diff_ops[line[0]]
    color = ops_color[op]
    return colorize_line_with_color(line, color, no_color)


def gen_pre_as_diff(
    pre: Dict,
    show_rules: bool,
    indent: str,
    no_color: bool,
    _level: int = 0
) -> Generator[str, None, None]:
    ops = [(order, op) for op, order in ops_order.items()]
    ops.sort()
    for (raw_rule, content) in pre.items():
        items = content["items"].items()
        for (_, diff) in items:  # pylint: disable=redefined-outer-name
            if show_rules and not raw_rule == "__MULTILINE_BODY__":
                line = "# %s%s\n" % (indent * _level, raw_rule)
                yield colorize_line_with_color(line, colorama.Fore.BLACK, no_color)
            for (op, rows) in [(op, diff[op]) for (_, op) in ops]:
                for item in rows:
                    line = "%s%s %s\n" % (ops_sign[op], indent * _level, item["row"])
                    yield colorize_line_with_color(line, ops_color[op], no_color)
                    if len(item["children"]) != 0:
                        yield from gen_pre_as_diff(item["children"], show_rules, indent, no_color, _level + 1)



import re

from annet.annlib.types import Op

from annet.rulebook import common


def permanent(rule, key, diff, **kwargs):  # pylint: disable=redefined-outer-name
    ifname = key[0]
    if re.match(r"(Eth-Trunk|Vlanif|Vbdif|Loop[Bb]ack|Tunnel|.*\.\d+)", ifname):
        # эти интерфейсы можно удалять
        yield from common.default(rule, key, diff, **kwargs)
    else:
        yield from common.permanent(rule, key, diff, **kwargs)


# [NOCDEV-2180] Хуавей просит переввести ip конфигурацию после изменения vrf
def binding_change(old, new, diff_pre, _pops=(Op.AFFECTED,)):
    ret = common.default_diff(old, new, diff_pre, _pops)
    vpn_changed = False
    for (op, cmd, _, _) in ret:
        if op in {Op.ADDED, Op.REMOVED}:
            vpn_changed |= _is_vpn_cmd(cmd)
    if vpn_changed:
        for cmd in list(old.keys()):
            if not _is_vpn_cmd(cmd):
                del old[cmd]
        ret = common.default_diff(old, new, diff_pre, _pops)
    return ret


def _is_vpn_cmd(cmd):
    return cmd.startswith("ip binding vpn-instance")



import os


class NetboxStorageOpts:
    def __init__(self, url: str, token: str):
        self.url = url
        self.token = token

    @classmethod
    def from_cli_opts(cls, cli_opts):
        return cls(
            url=os.getenv("NETBOX_URL", "http://localhost"),
            token=os.getenv("NETBOX_TOKEN", "").strip(),
        )



import contextlib
import inspect
from abc import ABC, abstractmethod
from functools import wraps
from typing import Optional, Type, Union

from annet.connectors import CachedConnector


MinDurationT = Optional[Union[str, int]]


class _Nop:
    def __getattr__(self, item):
        return _Nop()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def __call__(self, *args, **kwargs):
        return _Nop()

    def __bool__(self):
        return False


class Tracing(ABC):
    enabled = False

    @abstractmethod
    def function(self, arg=None, /, **kwargs):
        pass

    @abstractmethod
    def contextmanager(self, arg=None, /, **kwargs):
        pass

    @abstractmethod
    def get_current_span(self, context_value=None):
        pass

    @abstractmethod
    def set_device_attributes(self, span, device):
        pass

    @abstractmethod
    def start_as_current_span(self, *args, **kwargs):
        pass

    @abstractmethod
    def start_as_linked_span(self, *args, **kwargs):
        pass

    @abstractmethod
    def attach_context(self, *args, **kwargs):
        pass

    @abstractmethod
    def inject_context(self, *args, **kwargs):
        pass

    @abstractmethod
    def extract_context(self, *args, **kwargs):
        pass

    @abstractmethod
    def get_context(self, *args, **kwargs):
        pass

    @abstractmethod
    def force_flush(self):
        pass


class DummyTracing(Tracing):
    def function(self, arg=None, /, **kwargs):
        def decorator(func):
            return func

        return decorator if arg is None else decorator(arg)

    def contextmanager(self, arg=None, /, **kwargs):
        def decorator(cls_or_func):
            return cls_or_func

        return decorator if arg is None else decorator(arg)

    def get_current_span(self, context_value=None):
        return None

    def set_device_attributes(self, span, device):
        return None

    def start_as_current_span(self, *args, **kwargs):
        return _Nop()

    def start_as_linked_span(self, *args, **kwargs):
        return _Nop()

    def attach_context(self, *args, **kwargs):
        return

    def inject_context(self, *args, **kwargs):
        return

    def extract_context(self, *args, **kwargs):
        return

    def get_context(self, *args, **kwargs):
        return

    def force_flush(self):
        return


class _TracingConnector(CachedConnector[Tracing]):
    name = "Tracing"
    ep_name = "tracing"

    def _get_default(self) -> Type[Tracing]:
        return DummyTracing


tracing_connector = _TracingConnector()


def function(arg=None, /, **outer_kwargs):
    def decorator(func):
        cache = None

        @wraps(func)
        def wrapper(*args, **kwargs):
            nonlocal cache
            if cache is None:
                cache = tracing_connector.get().function(func, **outer_kwargs)
            return cache(*args, **kwargs)

        return wrapper

    return decorator if arg is None else decorator(arg)


def contextmanager(arg=None, /, **outer_kwargs):
    def decorator(cls_or_func):
        if inspect.isfunction(cls_or_func):
            cache = None

            @contextlib.contextmanager
            @wraps(cls_or_func)
            def wrapper(*args, **kwargs):
                nonlocal cache
                if cache is None:
                    cache = tracing_connector.get().contextmanager(cls_or_func, **outer_kwargs)
                with cache(*args, **kwargs) as val:
                    yield val

            return wrapper
        else:
            @wraps(cls_or_func.__enter__)
            def enter(*args, **kwargs):
                tracing_connector.get().contextmanager(cls_or_func, **outer_kwargs)  # redefine __enter__ and __exit__
                return cls_or_func.__enter__(*args, **kwargs)  # pylint: disable=unnecessary-dunder-call

            setattr(cls_or_func, "__enter__", enter)
            return cls_or_func

    return decorator if arg is None else decorator(arg)



from collections import OrderedDict
from typing import Any, Dict, List, MutableMapping, NamedTuple, Optional, Tuple, Union

from annet.annlib.types import Op  # pylint: disable=unused-import

from annet.storage import Device, Storage


class PCDiffFile(NamedTuple):
    label: str
    diff_lines: List[str]


class PCDiff(NamedTuple):
    hostname: str
    diff_files: List[PCDiffFile]


Diff = List[Tuple]
ExitCode = int


class GeneratorPerf:
    """
    Рантайм статистика времени выполнения генератора
    """

    def __init__(self, total: float, rt: float, meta: Optional[Dict[str, Any]] = None):
        self.total = total
        self.rt = rt
        self._meta = meta

    @property
    def meta(self) -> Dict[str, Any]:
        return self._meta or {}


class GeneratorPartialRunArgs:
    """
    Параметры и модификаторы для запуска run_partial_generators
    """

    def __init__(
        self,
        device: Device,
        use_acl: bool = False,
        use_acl_safe: bool = False,
        annotate: bool = False,
        generators_context: Optional[str] = None,
        no_new: bool = False,
    ):
        self.device = device
        self.use_acl = use_acl  # фильтруем по acl ввыод генератора (--no-acl для дебага)
        self.use_acl_safe = use_acl_safe  # [NOCDEV-6190] используем более строгий генераторный acl
        self.annotate = annotate  # добавляем в каждую строку вывода информацию откуда она была заyield'ена
        self.generators_context = generators_context  # строка с именем контекста генераторов
        self.no_new = no_new  # для опции --clear, не пытаемся запустить генераторы, выдаем только acl


class GeneratorPartialResult:
    """
    Результат запуска Partial-генератора
    """

    def __init__(
        self,
        name: str,
        tags: List[str],
        acl: str,
        acl_rules: Dict[str, Any],  # OrderedDict
        acl_safe: str,
        acl_safe_rules: Dict[str, Any],  # OrderedDict
        output: str,
        config: Dict[str, Any],  # OrderedDict
        safe_config: Dict[str, Any],  # OrderedDict
        perf: GeneratorPerf,
    ):
        self.name = name
        self.tags = tags
        self.acl = acl
        self.acl_rules = acl_rules
        self.acl_safe = acl_safe
        self.acl_safe_rules = acl_safe_rules
        self.output = output
        self.config = config
        self.safe_config = safe_config
        self.perf = perf


class GeneratorEntireResult:
    """
    Результат запуска Entire-генератора
    """

    def __init__(
        self,
        name: str,
        tags: List[str],
        path: Optional[str],
        output: str,
        reload: str,
        prio: int,
        perf: GeneratorPerf,
        is_safe: bool,
    ):
        self.name = name
        self.tags = tags
        self.path = path
        self.output = output
        self.reload = reload
        self.prio = prio
        self.perf = perf
        self.is_safe = is_safe


class GeneratorJSONFragmentResult:
    """
    Результат запуска JSONFragment-генератора
    """

    def __init__(
        self,
        name: str,
        tags: List[str],
        path: str,
        acl: List[str],
        acl_safe: List[str],
        config: Dict[str, Any],
        reload: str,
        perf: GeneratorPerf,
        reload_prio: int,
    ):
        self.name = name
        self.tags = tags
        self.path = path
        self.acl = acl
        self.acl_safe = acl_safe
        self.config = config
        self.reload = reload
        self.perf = perf
        self.reload_prio = reload_prio


GeneratorResult = Union[GeneratorEntireResult, GeneratorPartialResult, GeneratorJSONFragmentResult]


class OldNewResult:
    """Результат запуска old_new
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        device=None,
        old=None,
        new=None,
        acl_rules=None,
        new_files=None,
        old_files=None,
        err=None,
        partial_result=None,
        entire_result=None,
        old_json_fragment_files=None,
        new_json_fragment_files=None,
        json_fragment_result=None,
        implicit_rules=None,
        perf=None,
        acl_safe_rules=None,
        safe_old=None,
        safe_new=None,
        safe_new_files=None,
        safe_new_json_fragment_files=None,
        filter_acl_rules=None
    ):
        self.device: Device = device
        self.old: MutableMapping = old if old else OrderedDict()
        self.new: MutableMapping = new if new else OrderedDict()
        self.acl_rules: MutableMapping = acl_rules
        self.new_files: MutableMapping = new_files if new_files else {}
        self.old_files: MutableMapping = old_files if old_files else {}
        self.err: Optional[Exception] = err
        self.partial_results: Dict[str, GeneratorPartialResult] = partial_result or {}
        self.entire_results: Dict[str, GeneratorEntireResult] = entire_result or {}
        self.old_json_fragment_files: Dict[str, Any] = old_json_fragment_files or {}
        self.new_json_fragment_files: Dict[str, Tuple[Any, Optional[str]]] = new_json_fragment_files or {}
        self.json_fragment_results: Dict[str, GeneratorJSONFragmentResult] = json_fragment_result or {}
        self.implicit_rules: Dict[str, Any] = implicit_rules or OrderedDict()
        self.perf: Dict[str, Dict[str, float]] = perf or {}

        # safe acl and configs with it applied
        self.acl_safe_rules: MutableMapping = acl_safe_rules or {}
        self.safe_old: MutableMapping = safe_old if safe_old else OrderedDict()
        self.safe_new: MutableMapping = safe_new if safe_new else OrderedDict()
        self.safe_new_files: MutableMapping = safe_new_files if safe_new_files else {}
        self.safe_new_json_fragment_files: MutableMapping = safe_new_json_fragment_files or {}

        self.filter_acl_rules: Optional[MutableMapping] = filter_acl_rules

    def get_old(self, safe: bool = False) -> MutableMapping:
        if safe:
            return self.safe_old

        return self.old

    def get_new(self, safe: bool = False) -> MutableMapping:
        if safe:
            return self.safe_new

        return self.new

    def get_acl_rules(self, safe: bool = False) -> MutableMapping:
        if safe:
            return self.acl_safe_rules

        return self.acl_rules

    def get_new_files(self, safe: bool = False) -> MutableMapping:
        if safe:
            return self.safe_new_files

        return self.new_files

    def get_new_file_fragments(self, safe: bool = False) -> Dict[str, Tuple[Any, Optional[str]]]:  # pylint: disable=unused-argument
        if safe:
            return self.safe_new_json_fragment_files

        return self.new_json_fragment_files
